@@ -1,5 +1,5 @@
 /* function called by chkr_x_x_x_x_chkr, at syscalls.
-   Copyright 1993 Tristan Gingold
+   Copyright 1993, 1994, 1995 Tristan Gingold
 		  Written September 1993 Tristan Gingold
 
 This program is free software; you can redistribute it and/or
@@ -14,18 +14,15 @@ General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; see the file COPYING.  If not, write to
-the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.
 
- The author may be reached (Email) at Tristan C/O gingold@amoco.saclay.cea.fr,
- or (US/French mail) as Tristan Gingold
-   			  8 rue Parmentier
-   			  F91120 PALAISEAU
-   			  FRANCE
+The author may be reached by US/French mail:
+		Tristan Gingold 
+		8 rue Parmentier
+		F-91120 PALAISEAU
+		FRANCE
 */
-#define NEED_MM
-#include "checker.h"
-#include "errlist.h"
-#include "message.h"
 #include <stddef.h>
 #include <unistd.h>
 #include <utime.h>
@@ -44,12 +41,14 @@ the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <sys/utsname.h>
 #include <sys/mount.h>
 #include <sys/mman.h>
-#include <sys/vm86.h>
+#include <asm/vm86.h>
 #include <sys/ioctl.h>
+#include <sys/dirent.h>
 #include <sys/socketcall.h>
 #include <sys/mtio.h>
 #include <netinet/in.h>
 #define __KERNEL__		/* needed for msg, shm, sem */
+#include <linux/linkage.h>
 #include <sys/shm.h>
 #include <sys/sem.h>
 #include <sys/msg.h>
@@ -61,10 +60,17 @@ the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 #include <linux/fd.h>
 #include <linux/kd.h>
 #include <linux/vt.h>
+#include <linux/kernel.h>
 #include <linux/serial.h>
 #include <linux/route.h>
+#include <linux/netdevice.h>
 #include <linux/if_arp.h>
 
+#define NEED_MM
+#include "checker.h"
+#include "errlist.h"
+#include "message.h"
+
 struct reg
 {
   unsigned int eax;
@@ -76,23 +82,17 @@ struct reg
   unsigned int esi;
   unsigned int edi;
   unsigned int eip;
-  unsigned char flags;
+  unsigned int flags;
 };
 
 extern void make_syscall (void);
 extern void make_pid(void);
 extern void skip_syscall (void);
-extern void fork_symtab_file (void);
 extern void update_output_file(void);
-extern void exec_symtab_file (void);
-extern void remove_mmap(PTR addr, uint len);
-extern void new_segmmap(PTR addr, uint len, int prot, int flags, int filedes, uint off);
-extern void new_segshm(int shmid, PTR addr, int flags);
-extern void remove_shm(PTR addr);
 int check_ioctl (int cmd, int arg, int *const res);
 int check_ipc (ulong call, ulong first, ulong second, ulong third, void *ptr, int *res);
 int check_socketcall (int call, unsigned long *args, unsigned int *const res);
-int check_fcntl (int fd, int cmd, int arg);
+int check_fcntl (int fd, int cmd, int arg, int *const res);
 
 #define RES  ((signed long)(regs->eax))
 #define CALLNBR RES
@@ -109,35 +109,6 @@ int check_fcntl (int fd, int cmd, int arg);
 #define PTR4 (PTR)ARG4
 #define PTR5 (PTR)ARG5
 
-/* Check if the fd can be used by the user (fl = 0)
- * or can be returned by the system (fl = 1)
- */
-static inline int
-is_fd_ok (int fd, int fl)
-{
-  if (fd >= CHKR_FIRST_FD)
-    {
-      chkr_perror (fl ? M_I_SFD_SC_ET: M_I_BFD_SC_ET);
-      return 0;
-    }
-  return 1;
-}
-
-/* Check the address of socket, according to the protocol. */
-static inline void
-check_sockaddr(struct sockaddr *addr, int len)
-{
-    chkr_check_addr ((__ptr_t) addr, sizeof(short), CHKR_RO);
-    switch (addr->sa_family)
-      {
-    case AF_INET:
-	if (len > sizeof(short)+sizeof(short)+sizeof(struct in_addr))
-	  len = sizeof(short)+sizeof(short)+sizeof(struct in_addr);
-	break;
-    }
-    chkr_check_addr ((__ptr_t) addr, len, CHKR_RO);
-}
-
 /* This function is called by a stub to check a syscall.  All the registers 
  *  available through REGS and can be modified. */
 void
@@ -155,14 +126,10 @@ check_syscall (struct reg *const regs)
     case SYS_fork:
       make_syscall ();
       if (RES == 0)
-        {
-          make_pid();
-	  fork_symtab_file();
-	  update_output_file();
-	}
+        chkr_clean_after_fork ();
       break;
     case SYS_read:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       if (ARG3 > 0)
         chkr_check_addr (PTR2, ARG3, CHKR_TW);
       make_syscall ();
@@ -170,22 +137,24 @@ check_syscall (struct reg *const regs)
 	chkr_set_right (PTR2, RES, CHKR_RW);
       break;
     case SYS_write:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       if (ARG3 > 0)
         chkr_check_addr (PTR2, ARG3, CHKR_RO);
       break;
     case SYS_open:
       chkr_check_str (PTR1, CHKR_RO);
       make_syscall ();
-      if (RES != -1)
-	is_fd_ok (RES, 1);
+      if (RES >= 0)
+	fd_returned_by_system (RES);
       break;
     case SYS_close:
-      if (!is_fd_ok (ARG1, 0))
+      if (!fd_used_by_prog (ARG1))
 	{
 	  skip_syscall ();
 	  RES = -EBADF;
 	}
+      else
+        fd_closed(ARG1);
       break;
     case SYS_waitpid:
       error = 1;		/* replaced by wait4 */
@@ -229,8 +198,7 @@ check_syscall (struct reg *const regs)
 	  }
 	chkr_check_addr (ptr, sizeof (char *), CHKR_RO);
       }
-      chkr_do_end ();
-      exec_symtab_file ();
+      chkr_clean_before_exec ();
       break;
     case SYS_chdir:
       chkr_check_str (PTR1, CHKR_RO);
@@ -239,7 +207,11 @@ check_syscall (struct reg *const regs)
       if (ARG1)
 	chkr_check_addr (PTR1, sizeof (time_t), CHKR_WO);
       break;
+#ifdef SYS_prev_mknod
+    case SYS_prev_mknod:
+#else
     case SYS_mknod:
+#endif
       chkr_check_str (PTR1, CHKR_RO);
       break;
     case SYS_chmod:
@@ -255,7 +227,7 @@ check_syscall (struct reg *const regs)
       error = 1;
       break;
     case SYS_lseek:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       break;
     case SYS_getpid:
       break;
@@ -263,6 +235,7 @@ check_syscall (struct reg *const regs)
       chkr_check_str (PTR1, CHKR_RO);
       chkr_check_str (PTR2, CHKR_RO);
       chkr_check_str (PTR3, CHKR_RO);
+      /* FIXME: options */
       break;
     case SYS_umount:
       chkr_check_str (PTR1, CHKR_RO);
@@ -275,6 +248,7 @@ check_syscall (struct reg *const regs)
       chkr_check_addr (PTR1, sizeof (time_t), CHKR_RO);
       break;
     case SYS_ptrace:
+      /* FIXME: to do */
       error = 1;
       break;
     case SYS_alarm:
@@ -317,17 +291,17 @@ check_syscall (struct reg *const regs)
       chkr_check_str (PTR1, CHKR_RO);
       break;
     case SYS_dup:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       make_syscall ();
-      is_fd_ok (RES, 1);
+      fd_duped (RES, ARG1);
       break;
     case SYS_pipe:
       chkr_check_addr (PTR1, 2 * sizeof (int), CHKR_WO);
       make_syscall ();
       if (RES == 0)
 	{
-	  is_fd_ok (((int *) PTR1)[0], 1);
-	  is_fd_ok (((int *) PTR1)[1], 1);
+	  fd_returned_by_system (((int *) PTR1)[0]);
+	  fd_returned_by_system (((int *) PTR1)[1]);
 	}
       break;
     case SYS_times:
@@ -337,8 +311,9 @@ check_syscall (struct reg *const regs)
       error = 1;
       break;
     case SYS_brk:
+      /* FIXME: call brk ? */
       error = 1;
-      break;			/* called only by checker */
+      break;
     case SYS_setgid:
       break;
     case SYS_getgid:
@@ -360,12 +335,12 @@ check_syscall (struct reg *const regs)
       error = 1;
       break;
     case SYS_ioctl:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       error = check_ioctl (ARG2, ARG3, RESPTR);
       break;
     case SYS_fcntl:
-      /* NOTA: is_fd_ok is done inside check_fcntl */
-      error = check_fcntl (ARG1, ARG2, ARG3);
+      fd_used_by_prog (ARG1);
+      error = check_fcntl (ARG1, ARG2, ARG3, RESPTR);
       break;
     case SYS_mpx:
       error = 1;
@@ -383,15 +358,21 @@ check_syscall (struct reg *const regs)
     case SYS_chroot:
       chkr_check_str (PTR1, CHKR_RO);
       break;
+#ifdef SYS_prev_ustat
+    case SYS_prev_ustat:
+#else
     case SYS_ustat:
+#endif
       chkr_check_addr (PTR2, sizeof (struct ustat), CHKR_WO);
       break;
     case SYS_dup2:
-      if (!is_fd_ok (ARG1, 0) || !is_fd_ok (ARG2, 0))
+      if (!fd_used_by_prog (ARG1) || !fd_used_by_prog (ARG2))
 	{
 	  skip_syscall ();
 	  RES = -EBADF;
 	}
+      else
+        fd_duped(ARG1, ARG2);
       break;
     case SYS_getppid:
       break;
@@ -400,20 +381,22 @@ check_syscall (struct reg *const regs)
     case SYS_setsid:
       break;
     case SYS_sigaction:
-#if 1
-      __sigaction (ARG1, PTR2, PTR3);
-#else
-      if (ARG2)
-	chkr_check_addr (PTR2, sizeof (struct sigaction), CHKR_RO);
-      if (ARG3)
-	chkr_check_addr (PTR3, sizeof (struct sigaction), CHKR_WO);
-#endif
+      skip_syscall();
+      /* ACT is optional and is a pointer.  */
+      if (PTR2)
+        chkr_check_addr (PTR2, sizeof (struct sigaction), CHKR_RO);
+    
+      /* OLDACT is optional... */
+      if (PTR3)
+        chkr_check_addr (PTR3, sizeof (struct sigaction), CHKR_WO);
+  
+      RES = user_sigaction (ARG1, PTR2, PTR3);
       break;
     case SYS_siggetmask:
-      error = 1;		/* old syscall */
+      error = 1;		/* old syscall: sigprocmask */
       break;
     case SYS_sigsetmask:
-      error = 1;		/* old syscall */
+      error = 1;		/* old syscall: sigprocmask */
       break;
     case SYS_setreuid:
       break;
@@ -424,7 +407,7 @@ check_syscall (struct reg *const regs)
     case SYS_sigpending:
       break;
     case SYS_sethostname:
-      if (ARG2 > 0)
+      if (ARG2)
         chkr_check_addr (PTR1, ARG2, CHKR_RO);
       break;
     case SYS_setrlimit:
@@ -447,30 +430,68 @@ check_syscall (struct reg *const regs)
       chkr_check_addr (PTR2, sizeof (struct timezone), CHKR_RO);
       break;
     case SYS_getgroups:
-      if (ARG1 > 0)
+      if (ARG1)
         chkr_check_addr (PTR2, ARG1 * sizeof (__gid_t), CHKR_TW);
       make_syscall ();
       if (RES > 0)
 	chkr_set_right (PTR2, RES * sizeof (__gid_t), CHKR_RW);
       break;
     case SYS_setgroups:
-      if (ARG1 > 0)
+      if (ARG1)
         chkr_check_addr (PTR2, ARG1 * sizeof (__gid_t), CHKR_RO);
       break;
-    case SYS_select:		/* FIXME: is_fd_ok */
+    case SYS_select:
       {
 	char **ptr;
+	fd_set fd_used;
+	int i, j, nbr_fd;
 
+	/* Only one warning by fd ! */
+        FD_ZERO(&fd_used);
+        
 	chkr_check_addr (PTR1, 5 * sizeof (int *), CHKR_RO);
+	
+	/* Only the first NBR_FD are used */
+	nbr_fd = ((int *)ARG1)[0];
+	nbr_fd = nbr_fd > FD_SETSIZE ? FD_SETSIZE : nbr_fd;
+	
+	/* | could replace FD_SET, but the later one is more portable */
 	ptr = (char **) PTR1;
 	if (ptr[1])
-	  chkr_check_addr (ptr[1], sizeof (fd_set), CHKR_RW);
+	  {
+	    chkr_check_addr (ptr[1], sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, ptr[1]))
+	        FD_SET(i, &fd_used);
+	  }
 	if (ptr[2])
-	  chkr_check_addr (ptr[2], sizeof (fd_set), CHKR_RW);
+	  {
+	    chkr_check_addr (ptr[2], sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, ptr[2]))
+	        FD_SET(i, &fd_used);
+	  }
 	if (ptr[3])
-	  chkr_check_addr (ptr[3], sizeof (fd_set), CHKR_RW);
+	  {
+	    chkr_check_addr (ptr[3], sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, ptr[3]))
+	        FD_SET(i, &fd_used);
+	  }
 	if (ptr[4])
 	  chkr_check_addr (ptr[4], sizeof (struct timeval), CHKR_RW);
+	  
+	/* Now we can check each fd */
+	for (i = 0; i < nbr_fd; i++)
+	  if (FD_ISSET(i, &fd_used))
+	    if (!fd_used_by_prog(i))
+	      {
+	        /* This fd is reserved by the system, so remove it from 
+	         * the lists. */
+	        for (j = 1; j < 4; j++)
+	          if (ptr[j])
+	            FD_CLR(i, ptr[j]);
+	      }
       }
       break;
     case SYS_symlink:
@@ -482,7 +503,7 @@ check_syscall (struct reg *const regs)
       break;
     case SYS_readlink:
       chkr_check_str (PTR1, CHKR_RO);
-      if (ARG3 > 0)
+      if (ARG3)
         chkr_check_addr (PTR2, ARG3, CHKR_TW);
       make_syscall ();
       if (RES > 0)
@@ -497,7 +518,7 @@ check_syscall (struct reg *const regs)
     case SYS_reboot:
       break;
     case SYS_readdir:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       chkr_check_addr (PTR2, sizeof (struct dirent), CHKR_WO);
       break;
     case SYS_mmap:
@@ -505,7 +526,7 @@ check_syscall (struct reg *const regs)
         long *buffer = (long*)PTR1;
         chkr_check_addr(PTR1, 6 * sizeof(long int), CHKR_RO);
         if (!(buffer[3] & MAP_ANONYMOUS))
-          is_fd_ok(buffer[4], 0);
+          fd_used_by_prog(buffer[4]);
         if (buffer[0] >= MM_LOW && buffer[0] <= MM_HIGH)
           {
             if (buffer[3] & MAP_FIXED)
@@ -517,7 +538,7 @@ check_syscall (struct reg *const regs)
             buffer[0] = 0;
           }
         make_syscall();
-        if (RES < 0 && RES > -1000)     /* up to 1000 errors ! FIXME */
+        if (RES < 0 && RES > -4096)     /* up to 1000 errors ! FIXME */
           break;        /* fail */
         new_segmmap((void*)RES, buffer[1], buffer[2], buffer[3],
                     buffer[4], buffer[5]);
@@ -537,14 +558,15 @@ check_syscall (struct reg *const regs)
     case SYS_truncate:
       chkr_check_str (PTR1, CHKR_RO);
       break;
+
     case SYS_ftruncate:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       break;
     case SYS_fchmod:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       break;
     case SYS_fchown:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       break;
     case SYS_getpriority:
       break;
@@ -558,7 +580,7 @@ check_syscall (struct reg *const regs)
       chkr_check_addr (PTR2, sizeof (struct statfs), CHKR_WO);
       break;
     case SYS_fstatfs:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       chkr_check_addr (PTR2, sizeof (struct statfs), CHKR_WO);
       break;
     case SYS_ioperm:
@@ -567,19 +589,27 @@ check_syscall (struct reg *const regs)
       error = check_socketcall (ARG1, (unsigned long *) ARG2, RESPTR);
       break;
     case SYS_klog:
-      if (ARG1 == 0 || ARG1 == 1 || ARG1 == 5
-	  || ARG1 == 6 || ARG1 == 7 || ARG1 == 8)
-	break;			/* nothing to check */
-      else if (ARG1 == 2 || ARG1 == 3 || ARG1 == 4)
-	{
-	  if (ARG3 > 0)
-	    chkr_check_addr (PTR2, ARG3, CHKR_TW);
-	  make_syscall ();
-	  if (RES > 0)
-	    chkr_set_right (PTR2, RES, CHKR_RW);
-	}
-      else
-	error = 1;
+      switch (ARG1)
+        {
+          case 0:
+          case 1:
+          case 5:
+          case 6:
+          case 7:
+          case 8:
+            break;			/* nothing to check */
+          case 2:
+          case 3:
+          case 4:
+            if (ARG3)
+              chkr_check_addr (PTR2, ARG3, CHKR_TW);
+            make_syscall ();
+            if (RES > 0)
+              chkr_set_right (PTR2, RES, CHKR_RW);
+            break;
+          default:
+            error = 1;
+        }
       break;
     case SYS_setitimer:
       chkr_check_addr (PTR2, sizeof (struct itimerval), CHKR_RO);
@@ -589,16 +619,28 @@ check_syscall (struct reg *const regs)
     case SYS_getitimer:
       chkr_check_addr (PTR2, sizeof (struct itimerval), CHKR_WO);
       break;
+#ifdef SYS_prev_stat
+    case SYS_prev_stat:
+#else
     case SYS_stat:
+#endif
       chkr_check_str (PTR1, CHKR_RO);
       chkr_check_addr (PTR2, sizeof (struct stat), CHKR_WO);
       break;
+#ifdef SYS_prev_lstat
+    case SYS_prev_lstat:
+#else
     case SYS_lstat:
+#endif
       chkr_check_str (PTR1, CHKR_RO);
       chkr_check_addr (PTR2, sizeof (struct stat), CHKR_WO);
       break;
+#ifdef SYS_prev_fstat
+    case SYS_prev_fstat:
+#else
     case SYS_fstat:
-      is_fd_ok (ARG1, 0);
+#endif
+      fd_used_by_prog (ARG1);
       chkr_check_addr (PTR2, sizeof (struct stat), CHKR_WO);
       break;
     case SYS_olduname:
@@ -623,20 +665,23 @@ check_syscall (struct reg *const regs)
       chkr_check_str (PTR1, CHKR_RO);
       break;
     case SYS_sysinfo:
-      error = 1;
+      /* 22 is due to the pad */
+      chkr_check_addr (PTR1, sizeof (struct sysinfo) - 22, CHKR_WO);
       break;
     case SYS_ipc:
       error = check_ipc (ARG1, ARG2, ARG3, ARG4, PTR5, RESPTR);
       break;
     case SYS_fsync:
-      is_fd_ok (ARG1, 0);
+      fd_used_by_prog (ARG1);
       break;
     case SYS_sigreturn:
       break;
     case SYS_clone:
+      /* Not supported because of the stack and recursion */
+      error = 1;
       break;
     case SYS_setdomainname:
-      if (ARG2 > 0)
+      if (ARG2)
         chkr_check_addr (PTR1, ARG2, CHKR_RO);
       break;
     case SYS_uname:
@@ -677,12 +722,11 @@ check_syscall (struct reg *const regs)
     case SYS_fchdir:
       break;
     case SYS_bdflush:
-      error = 1;
+      if (ARG1 >= 2 && (ARG1 & 1) == 0)
+        chkr_check_addr(PTR2, CHKR_WO, sizeof (int));
       break;
-    /* If you don't have SYS_sysfs or later, either get a more recent kernel
-       or uncomment the next comments. */
-/* #if 0 */ /* To be uncommented if necessary */       
     case SYS_sysfs:
+      /* Too weird */
       error = 1;
       break;
     case SYS_personality:
@@ -691,13 +735,91 @@ check_syscall (struct reg *const regs)
     case SYS_afs_syscall:
       error = 1;
       break;
-/* #endif */ /* To be uncommented if necessary */      
+#if 0
+    case SYS_setfsuid:
+      break;
+    case SYS_setfsgid:
+      break;
+    case SYS__llseek:
+      fd_used_by_prog(ARG1);
+      chkr_check_addr(PTR4, CHKR_WO, sizeof(loff_t));
+      break;
+#endif
+    case SYS_getdents:
+      fd_used_by_prog (ARG1);
+      if (ARG3 < 0)
+	error = 1;
+      else
+	chkr_check_addr (PTR2, ARG3, CHKR_WO);
+      break;
+    case SYS__newselect:
+      {
+	fd_set fd_used;
+	int i, j, nbr_fd;
+
+	/* Only one warning by fd ! */
+        FD_ZERO(&fd_used);
+        
+	/* Only the first NBR_FD are used */
+	nbr_fd = ARG1;
+	nbr_fd = nbr_fd > FD_SETSIZE ? FD_SETSIZE : nbr_fd;
+	
+	/* | could replace FD_SET, but the later one is more portable */
+	if (PTR2)
+	  {
+	    chkr_check_addr (PTR2, sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, PTR2))
+	        FD_SET(i, &fd_used);
+	  }
+	if (PTR3)
+	  {
+	    chkr_check_addr (PTR3, sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, PTR3))
+	        FD_SET(i, &fd_used);
+	  }
+	if (PTR4)
+	  {
+	    chkr_check_addr (PTR4, sizeof (fd_set), CHKR_RW);
+	    for (i = 0; i < nbr_fd; i++)
+	      if (FD_ISSET(i, PTR4))
+	        FD_SET(i, &fd_used);
+	  }
+	if (PTR5)
+	  chkr_check_addr (PTR5, sizeof (struct timeval), CHKR_RW);
+	  
+	/* Now we can check each fd */
+	for (i = 0; i < nbr_fd; i++)
+	  if (FD_ISSET(i, &fd_used))
+	    if (!fd_used_by_prog(i))
+	      {
+		fd_set *ptr[] = {0, PTR2, PTR3, PTR4, 0};
+		  
+	        /* This fd is reserved by the system, so remove it from 
+	         * the lists. */
+	        for (j = 1; j < 4; j++)
+	          if (ptr[j])
+	            FD_CLR(i, ptr[j]);
+	      }
+      }
+      break;
+    case SYS_nanosleep:
+      if (PTR1)
+	chkr_check_addr (PTR1, sizeof (struct timespec), CHKR_RO);
+      if (PTR2)
+	chkr_check_addr (PTR2, sizeof (struct timespec), CHKR_TW);
+      make_syscall ();
+      if (RES == -1)
+	chkr_set_right (PTR2, sizeof (struct timespec), CHKR_RW);
+      break;
     default:
       error = 1;
     }
   if (error)
     {
       chkr_perror (M_I_IES_SC_ET);
+      chkr_printf ("Don't know syscall %d\n", (int) CALLNBR);
       chkr_abort ();
     }
 }
@@ -739,7 +861,7 @@ check_ioctl (int cmd, int arg, int *const res)
       break;
     case TCXONC:		/* Mon Apr 4 1994 */
     case TCFLSH:		/* Mon Apr 4 1994 */
-      chkr_check_addr ((PTR) arg, sizeof (int), CHKR_RO);
+      /*chkr_check_addr ((PTR) arg, sizeof (int), CHKR_RO);*/
       break;
     case TIOCEXCL:		/* Mon Apr 4 1994 */
     case TIOCNXCL:		/* Mon Apr 4 1994 */
@@ -1139,13 +1261,13 @@ check_ioctl (int cmd, int arg, int *const res)
 }
 
 int
-check_fcntl (int fd, int cmd, int arg)
+check_fcntl (int fd, int cmd, int arg, int *const res)
 {
-  is_fd_ok (fd, 0); 
   switch (cmd)
     {
     case F_DUPFD:		/* arg is used as a value */
-      is_fd_ok (arg, 0);
+      make_syscall();
+      fd_duped(*res, fd);
       break;
     case F_GETFD:
     case F_GETFL:
@@ -1156,7 +1278,7 @@ check_fcntl (int fd, int cmd, int arg)
       chkr_check_addr ((PTR) arg, sizeof (struct flock), CHKR_WO);
       break;
     case F_SETLK:
-    case F_SETLKW:		/* TG: what is the difference ? */
+    case F_SETLKW:	/* Wait */
       chkr_check_addr ((PTR) arg, offsetof(struct flock,l_pid), CHKR_RO);
       break;
     case F_SETOWN:		/* arg is used as a value */
@@ -1170,6 +1292,21 @@ check_fcntl (int fd, int cmd, int arg)
   return 0;
 }
 
+/* Check the address of socket, according to the protocol. */
+static inline void
+check_sockaddr(struct sockaddr *addr, int len)
+{
+    chkr_check_addr ((__ptr_t) addr, sizeof(short), CHKR_RO);
+    switch (addr->sa_family)
+      {
+    case AF_INET:
+	if (len > sizeof(short)+sizeof(short)+sizeof(struct in_addr))
+	  len = sizeof(short)+sizeof(short)+sizeof(struct in_addr);
+	break;
+    }
+    chkr_check_addr ((__ptr_t) addr, len, CHKR_RO);
+}
+
 int
 check_socketcall (int call, unsigned long *args, unsigned int *const res)
 {
@@ -1177,7 +1314,7 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
   switch (call)
     {
     case SYS_ACCEPT:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       chkr_check_addr ((PTR) args[2], sizeof (int), CHKR_RO);
       if (*(int *) args[2] > 0)
         chkr_check_addr ((PTR) args[1], *(int *) args[2], CHKR_TW);
@@ -1186,17 +1323,17 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
 	chkr_set_right ((PTR) args[1], *(int *) args[2], CHKR_RW);
       break;
     case SYS_BIND:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         check_sockaddr((struct sockaddr*) args[1], args[2]);
       break;
     case SYS_CONNECT:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         check_sockaddr((struct sockaddr*) args[1], args[2]);
       break;
     case SYS_GETPEERNAME:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       chkr_check_addr ((PTR) args[2], sizeof (int), CHKR_RW);
       if (*(int*)args[2] > 0)
         chkr_check_addr ((PTR) args[1], *(int *) args[2], CHKR_TW);
@@ -1205,7 +1342,7 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
 	chkr_set_right ((PTR) args[1], *(int *) args[2], CHKR_RW);
       break;
     case SYS_GETSOCKNAME:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       chkr_check_addr ((PTR) args[2], sizeof (int), CHKR_RW);
       if (*(int*)args[2] > 0)
         chkr_check_addr ((PTR) args[1], *(int *) args[2], CHKR_TW);
@@ -1214,7 +1351,7 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
 	chkr_set_right ((PTR) args[1], *(int *) args[2], CHKR_RW);
       break;
     case SYS_GETSOCKOPT:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       chkr_check_addr ((PTR) args[4], sizeof (int), CHKR_RW);
       if (*(int*)args[4] > 0)
         chkr_check_addr ((PTR) args[3], *(int *) args[4], CHKR_TW);
@@ -1223,10 +1360,10 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
 	chkr_set_right ((PTR) args[3], *(int *) args[4], CHKR_RW);
       break;
     case SYS_LISTEN:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       break;
     case SYS_RECV:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         chkr_check_addr ((PTR) args[1], args[2], CHKR_TW);
       make_syscall ();
@@ -1234,7 +1371,7 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
 	chkr_set_right ((PTR) args[1], *res, CHKR_RW);
       break;
     case SYS_RECVFROM:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         chkr_check_addr ((PTR) args[1], args[2], CHKR_TW);
       if (args[4])
@@ -1246,44 +1383,43 @@ check_socketcall (int call, unsigned long *args, unsigned int *const res)
       make_syscall ();
       if (*res > 0)
 	{
-	  if (args[2] > 0)
-	    chkr_set_right ((PTR) args[1], args[2], CHKR_RW);
+          chkr_set_right ((PTR) args[1], *res, CHKR_RW);
 	  if (args[4] > 0)
 	    chkr_set_right ((PTR) args[4], *(int *) args[5], CHKR_RW);
 	}
       break;
     case SYS_SEND:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         chkr_check_addr ((PTR) args[1], args[2], CHKR_RO);
       break;
     case SYS_SENDTO:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       if (args[2] > 0)
         chkr_check_addr ((PTR) args[1], args[2], CHKR_RO);
       if (args[5] > 0)
         check_sockaddr((struct sockaddr*) args[4], args[5]);
       break;
     case SYS_SETSOCKOPT:
-      is_fd_ok (args[0], 0);
-      if (args[5] > 0)
-        chkr_check_addr ((PTR) args[4], args[5], CHKR_RO);
+      fd_used_by_prog (args[0]);
+      if (args[4] > 0)
+        chkr_check_addr ((PTR) args[3], args[4], CHKR_RO);
       break;
     case SYS_SHUTDOWN:
-      is_fd_ok (args[0], 0);
+      fd_used_by_prog (args[0]);
       break;
     case SYS_SOCKET:
       make_syscall ();
       if (*res != -1)
-	is_fd_ok (*res, 1);
+	fd_returned_by_system (*res);
       break;
     case SYS_SOCKETPAIR:
       chkr_check_addr ((PTR) args[3], 2 * sizeof (int), CHKR_WO);
       make_syscall ();
       if (*res != -1)
 	{
-	  is_fd_ok (((int *) args[3])[0], 1);
-	  is_fd_ok (((int *) args[3])[1], 1);
+	  fd_returned_by_system (((int *) args[3])[0]);
+	  fd_returned_by_system (((int *) args[3])[1]);
 	}
       break;
     default:
